#!/usr/bin/perl -w
# -----------------------------------------------------------------------------
#
# addjailuser.pl
# Add user to the Chrooted environment
# A member of Jail Chroot Project.
# Released under GPL license version 2.0 
# 
# Juan M. Casillas
# assman@gsyc.inf.uc3m.es
#
# This script creates the required files and directories in the chrooted
# environment, edit the passwd / group / shadow files, and change the
# permissions for a given user
#
# -----------------------------------------------------------------------------

$CREATE_DIR_MASK = 0755;

# -----------------------------------------------------------------------------
#
# get_entry(file,uid)
# get the entry for a passwd, shadow or group file
#
# -----------------------------------------------------------------------------

sub get_entry {
  local $filename = $_[0];
  local $uid = $_[1];

  open(PASS,$filename) || return(0);
  while (<PASS>) {
    #
    # if matched, send it to the right file
    #
    if (/^$uid:/) {
      $entry = $_;
      last;
    }
  }

close(PASS);
return($entry);

}

# -----------------------------------------------------------------------------
#
# add_to_file(file,line)
# add line to the end of the file
#
# -----------------------------------------------------------------------------

sub add_to_file {
  local $filen = $_[0];
  local $linev = $_[1];
  
  open(FILEH,">>".$filen) || return(0);
  print FILEH $linev;
  close(FILEH);
  
  return(1);
}


# -----------------------------------------------------------------------------
#
# mkdir_recursive()
# create a directory recursively (works as mkdir -p)
#
# -----------------------------------------------------------------------------

sub mkdir_recursive {

 local $dst_dir = $_[0];

 local $createdir = "/";
 foreach  $i (split(/\//,$dst_dir)) {

   if ($i ne "") {
     $createdir = $createdir.$i;
     mkdir($createdir,$CREATE_DIR_MASK);
     
     $createdir = $createdir."/";
   }
 }

 return(1);
}


# -----------------------------------------------------------------------------
#
# gen_user(basedir,userid)
# generate the user directory inside de chrooted environment
#
# -----------------------------------------------------------------------------

sub gen_user {
 local $basedir = $_[0];
 local $userid  = $_[1];
 local $userhome = $_[2];
 local $usershell = $_[3];

 local $pass_file = "$basedir/etc/passwd";
 local $grp_file = "$basedir/etc/group";
 local $shw_file = "$basedir/etc/shadow";

 #
 # edit the passwd file
 #

 $entry = get_entry("/etc/passwd",$userid);
 if (!$entry) {
   return(0);
 }

 @values = split(/:/,$entry);
 $values[5] = $userhome;
 $values[6] = $usershell;

 $user_uid = $values[2];
 $user_gid = $values[3];
 
 $entry2 = join(':',@values)."\n";

 add_to_file($pass_file,$entry2) || return(0);

 #
 # build the group and the shadow file
 #

 $entry = get_entry("/etc/group",$userid) || return(0);
 add_to_file($grp_file,$entry);
 
 $entry = get_entry("/etc/shadow",$userid);
 add_to_file($shw_file,$entry);

 #
 # generate home directory
 # and change its permissions
 #

 mkdir_recursive("$basedir/$userhome") || return(0);
 chown($user_uid, $user_gid, "$basedir/$userhome") || return(0);

 return(1);
}

# -----------------------------------------------------------------------------
#
# show_credits()
# show basic credits
#
# -----------------------------------------------------------------------------

sub show_logo {

print "addjailuser.pl\n";
print "A component of Jail\n";
print "http://www.gsyc.inf.uc3m.es/~assman/jail/\n";
print "Juan M. Casillas <assman\@gsyc.inf.uc3m.es>\n\n";

}

# -----------------------------------------------------------------------------
#
# show_usage()
# show basic usage
#
# -----------------------------------------------------------------------------

sub show_usage {
  
  local $prg_name = $_[0];
 
show_logo();
print "usage: $prg_name chrootdir userdir usershell username when:\n\n";
 
  print "chrootdir       is the directory to chroot (e.g. /tmp/chroot)\n";
  print "userdir         is the user home directory in the chrooted\n";
  print "                environment (e.g. /home/user1)\n";
  print "usershell       is the full path to the users' shell\n";
  print "                (e.g. /bin/bash)\n";
  print "username        the username (e.g. user1)\n\n";

}

# -----------------------------------------------------------------------------
#
# main loop
#
# -----------------------------------------------------------------------------

if ( $#ARGV + 1 <  4 ) {
  show_usage($0);
  exit(0);
}

$chroot_dir = $ARGV[0];
$user_dir  = $ARGV[1];
$user_shell = $ARGV[2];
$user_name = $ARGV[3];

show_logo();
print "Adding user $user_name in  chrooted environment $chroot_dir\n";
if (!gen_user($chroot_dir,$user_name,$user_dir,$user_shell)) {
  print("\tError: Can't add the user.\n");
}
print "Done.\n";
