///////////////////////////////////////////////////////////////////////////////
//
// Jail
// A chrooted login shell implementation
// Released under GNU Public License v.2.0
// See LICENSE file for details
//
// Juan M. Casillas <assman@gsyc.inf.uc3m.es>
// 3/April/ 2001 09:22
// 
// Last Version: 1.8
//
///////////////////////////////////////////////////////////////////////////////
//
// $Id: jail.c,v 1.8 2001/05/07 11:29:42 assman Exp $
//
// $Log: jail.c,v $
// Revision 1.8  2001/05/07 11:29:42  assman
// erased #define SHELL due it not user any more
//
// Revision 1.7  2001/04/16 08:47:04  assman
// fixed a memory leak bug, fixed the directory read bug also
//
// Revision 1.5  2001/04/10 15:42:13  assman
// added documentation, change the Makefile to check de dependencies
// with mkenv.sh added TODO file, removed debug messages in the code
//
// Revision 1.4  2001/04/10 10:33:23  assman
// added CHANGELOG, INSTALL files, moved mkenv.sh for each platform,
// added new info to the documentation
//
// Revision 1.2  2001/04/03 10:28:36  assman
// removed binary file added by error
// added CVS tags
//
// 11/April/2001 --  I lost the cvs in my portable PC :(
//                   We are going to do a major cleanup in the code
//                   and improve some parts of Jail
//
// 13/April/2001 --  The user can choose the shell he wants using the
//                   "chrooted" /etc/passwd file. Major improvements in
//                   mkenv.sh, now it supports guessing the libraries
//                   it will use (we need to fine-tune this process using
//                   strace, this program is not installed on my
//                   portable PC :( )
//
// 05/July/2001  --  Fixed a segmentation fault error when trying to
//                   execute jail form the shell (there are not valid
//                   jail environment created in a valid user account)
//
///////////////////////////////////////////////////////////////////////////////
//
// NOTES
//
// File must be owned by root and have the SUID permissions
// (4755 runs fine) to allow proper operation. Also the chroot
// environment must be setup right. Se documentation files (README, INSTALL)
// for configuration and instalation
//
// ABOUT SSH & SCP
//
// scp passess the following parameters to Jail:
//
// jail -c scp -t /tmp
//
// from the scp's manual
//
// -t exit after reading and executing one command
// -c string       If  the  -c option is present, then commands are
//                 read from string.  If there are arguments  after
//                 the  string, they are assigned to the positional
//                 parameters, starting with $0.
//
//
// the solution is to pass all the parameters that are passing to jail to 
// the subshell, so the programs start to run. Also, we need to manage the
// terminal devices for proper work
//
///////////////////////////////////////////////////////////////////////////////
//
// BUGS
//
// some curious happen when trying to run jail in an IRIX 6.5
// when it starts, it becomes killed by itself. Tracing it, I
// think is a problem with the execve() system call. 
// Work arrounds for this platform are wellcome.
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>

#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <errno.h>
#include <grp.h>


//////////////////////////////////////////////////////////////////////////////
//
// DEFINE section
//
// VERSION           The CVS version of the program. This is not the 
//                   release tag, is the revision tag. You shouldn't change
//                   it
//
// HOME_VAR          When building user's homedir path, we need to prefix
//                   the HOME variable with this header.
//                 
// SHELL_VAR         When building user's shell variable, we need to prefix
//                   the SHELL variable with this header.
//
// BANNER            Some info about Jail program
//
//////////////////////////////////////////////////////////////////////////////


#define VERSION "$Id: jail.c,v 1.8 2001/05/07 11:29:42 assman Exp $"

#define BANNER "\nJail version " VERSION "\n\n" \
	       "This program is released under the GNU License\n" \
	       "Juan M. Casillas <assman@gsyc.inf.uc3m.es>\n" \
	       "See http://www.gsyc.inf.uc3m.es/~assman/jail\n" \
	       "for more info\n\n"

#define HOME_VAR "HOME="
#define SHELL_VAR "SHELL="
#define TERMINAL_GROUP "tty"

//////////////////////////////////////////////////////////////////////////////
//
// init_env()
// Duplicate a string and save it into the pointer pointed by ptr
//
// INPUT args:
//
//       **ptr        the pointer that will point to the dup string
//       *str         a pointer to the string
//
// OUTPUT args:
//
//       noting
//
//////////////////////////////////////////////////////////////////////////////

void init_env(char **ptr, char *str) {

    *ptr = (char *)malloc(strlen(str)+1);
    memset(*ptr,0,strlen(str)+1);
    strcpy(*ptr,str);
}

//////////////////////////////////////////////////////////////////////////////
//
// device_data
// a struct to handle the data about a terminal device
//
//       *path:     the path name of the device
//       major:     major number
//       minor:     minor number
//
//////////////////////////////////////////////////////////////////////////////

struct device_data {
    char *path;
    int major;
    int minor;
};    

//////////////////////////////////////////////////////////////////////////////
//
// generate_env()
// Create the terminal device.
//
// INPUT args:
//
//       *chroot:   path to the chrooted environment
//       *data:     the path name of the device, and the major and minor 
//       uid:       user id 
//       gid:       group id
//  
// OUTPUT args:
//
//       noting
//
//////////////////////////////////////////////////////////////////////////////

void generate_dev(char *chroot, struct device_data *data, int uid, int gid) {
    char *full_name;
    int sz;

    sz = strlen(chroot)+strlen(data->path)+2;

    full_name = (char *)malloc(sz);
    bzero(full_name,sz);
    strcpy(full_name,chroot);
    strcat(full_name,"/");
    strcat(full_name,data->path);

    mknod(full_name,S_IFCHR, makedev(data->major,data->minor));
    chown(full_name,uid,gid);
    chmod(full_name, S_IRUSR | S_IWUSR | S_IWGRP); // 620

    free(data);
    free(full_name);
}

//////////////////////////////////////////////////////////////////////////////
//
// get_tid_info()
// get the terminal info (path, major and minor numbers here)
//
// INPUT args:
//
//      nothing
//  
// OUTPUT args:
//
//      NULL           if the terminal isn't found
//      device_data*   a pointer to a struct with the data
//
//////////////////////////////////////////////////////////////////////////////

struct device_data *get_tid_info(void) {

    struct stat st_data;
    char *term_tid;
    
    struct device_data *ddata;

    term_tid = ctermid(NULL);
    if (term_tid == NULL)
	return(NULL);

    stat(term_tid, &st_data);
    if (!S_ISCHR(st_data.st_mode)) 
	return(NULL);

    ddata = (struct device_data *)malloc(sizeof(struct device_data));
    ddata->path=(char *)malloc(strlen(term_tid)+1);
    bzero(ddata->path,strlen(term_tid)+1);
    strcpy(ddata->path,term_tid);
    
    ddata->major = major(st_data.st_rdev);
    ddata->minor = minor(st_data.st_rdev);

    return(ddata);
}


//////////////////////////////////////////////////////////////////////////////
//
// get_tty_info()
// get the terminal info (path, major and minor numbers here)
//
// INPUT args:
//
//      nothing
//  
// OUTPUT args:
//
//      NULL           if the terminal isn't found
//      device_data*   a pointer to a struct with the data
//
//////////////////////////////////////////////////////////////////////////////

struct device_data *get_tty_info(void) {

    struct stat st_data;
    char *term_tty;
    
    struct device_data *ddata;

    term_tty = ttyname(0);

    if (term_tty == NULL)
	return(NULL);

    stat(term_tty, &st_data);
    if (!S_ISCHR(st_data.st_mode)) 
	return(NULL);

    ddata = (struct device_data *)malloc(sizeof(struct device_data));
    ddata->path=(char *)malloc(strlen(term_tty)+1);
    bzero(ddata->path,strlen(term_tty)+1);
    strcpy(ddata->path,term_tty);
    
    ddata->major = major(st_data.st_rdev);
    ddata->minor = minor(st_data.st_rdev);

    return(ddata);
}

//////////////////////////////////////////////////////////////////////////////
//
// build_directory()
// create a directory under the chrooted environment with permissions 755
//
// INPUT args:
//
//      
// OUTPUT args:
//
//       nothing
//
//////////////////////////////////////////////////////////////////////////////

void build_directory(char *chroot, char *dir) {
    char *full_name;
    int sz;
    
    sz = strlen(chroot)+strlen(dir)+2;
    full_name = (char *)malloc(sz);
    bzero(full_name,sz);

    strcpy(full_name,chroot);
    strcat(full_name,"/");
    strcat(full_name,dir);

    
    mkdir(full_name,0755);
    chmod(full_name,S_IRUSR|S_IXUSR|S_IWUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);

    free(full_name);
}


//////////////////////////////////////////////////////////////////////////////
//
// main()
// main execution loop
// 
// INPUT args:
//
//       argc           an integer that counts the number of arguments
//       *argv[]        an array of pointers pointing to the arguments
//       *env[]         an array of pointers pointing to the enviroment 
//                      variables passed from the caller program (usually, 
//                      a login daemon, just like telnetd, sshd, etc).
//
// OUTPUT args:
// 
//       Jail nevers returns (is overwritten by another process)
//       if some error ocurrs then the following values are returned:
//
//             
//             2        Errno in the following cases:
// 
//                      1) can't do a chdir()
//                      2) can't do a chroot()
//                      3) can't do a execve()
//             
//            -1        If we reached the last line of jail, that was a 
//                      really big problem, because jail should never 
//                      returns
//
//////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[], char *env[]) {
    int ret;
    char **lenv;
    char **larg;
    int n;
    struct passwd *pwdent,*cpwdent;
    int uid, gid;
    char *home_dir;
    char *shell_path;
    char *ptr_home_dir;
    int argc_cnt;
    int terminal_gid;

    struct group *tty_group_info;

    struct device_data *dev_tty, *dev_tid;

    //
    // here we have the full terminal device
    // so get the information about the device here
    //

    dev_tid = get_tid_info();
    dev_tty = get_tty_info();

    //
    // Get user id and group id
    // become root
    // get passwd file info for the user
    //

    uid = getuid();
    gid = getgid();

    if (setgid(0) || setuid(0) || getuid() != geteuid() || 
	getgid() != getegid()) {
	fprintf(stderr,"I can't do the chroot() because I'm not root\n\n");
	exit(-1);
    }

    //
    // save the chroot pointer for later
    //

    pwdent = getpwuid(uid);	

    //
    // create the terminal devices here
    //

    tty_group_info = getgrnam(TERMINAL_GROUP);
    if (tty_group_info == NULL)
	terminal_gid = gid;
    else
	terminal_gid = tty_group_info->gr_gid;

    if (dev_tid != NULL) {
	build_directory(pwdent->pw_dir,"dev");
	generate_dev(pwdent->pw_dir,dev_tid,uid,terminal_gid);
    }

    if (dev_tty != NULL) {
	build_directory(pwdent->pw_dir,"dev/pts");
	generate_dev(pwdent->pw_dir,dev_tty,uid,terminal_gid);
    }

    //
    // count elements and generate space for the array
    // then, save all the environment variables but do
    // nothing with HOME and SHELL
    //

    n=0;
    while (env[n]!=NULL) n++;  
    n++;

    lenv = malloc(sizeof(char *)*n);
    bzero(lenv,n);

    n=argc_cnt=0;
    while (env[n]) {
	if ((strstr(env[n],"HOME")==NULL) && 
	    (strstr(env[n],"SHELL")==NULL)) {
	    init_env(&lenv[argc_cnt],env[n]); 
	    //printf("%s\n",lenv[j]);
	    argc_cnt++;
	}
	n++;
    }	
    
    //
    // chdir & Chroot the users' directory readed before in pwent
    // Read again the passwd file info, that will be /virtual/etc/passwd
    // if the two entries are the same, do nothing, else, chdir to the
    // new directory, and save this info in the new environment argument
    // array.
    // 
    // next, get the shell entry from the chrooted /etc/passwd file
    // and set it as shell variable
    //

    ret = chdir(pwdent->pw_dir);	
    if (ret) {
	perror("chdir() ");
	return(ret);
    }
	
    ret = chroot(pwdent->pw_dir);
    if (ret) { 
	perror("chroot() ");
	return(ret);
    }

    cpwdent = getpwuid(uid);

    if (cpwdent == NULL) {
	printf("chrooted directory %s is not configured for jail; "
	       "bailing out\n",pwdent->pw_dir);
	return(-1);
    }


    if (strcmp(pwdent->pw_dir,cpwdent->pw_dir)) 
	ptr_home_dir = cpwdent->pw_dir;
    else
	ptr_home_dir = pwdent->pw_dir;

    n = strlen(HOME_VAR)+strlen(ptr_home_dir)+1;
    home_dir = (char *)malloc(n);
    memset(home_dir,0,n);
    strcpy(home_dir,HOME_VAR);
    strcat(home_dir,ptr_home_dir);

    n = strlen(SHELL_VAR)+strlen(cpwdent->pw_shell)+1;
    shell_path = (char *)malloc(n);
    memset(shell_path,0,n);
    strcpy(shell_path,SHELL_VAR);
    strcat(shell_path,cpwdent->pw_shell);


    init_env(&lenv[argc_cnt],home_dir); argc_cnt++;


    init_env(&lenv[argc_cnt],shell_path); argc_cnt++;	
    lenv[argc_cnt]=NULL;

    //
    // if we have some args in the call of jail, pass them to the
    // chrooted shell
    //

    if (argc > 0) {
	n=0;
	while(argv[n]) n++;  
	n++;

	larg = malloc(sizeof(char *)*n);	

	n = 0 ;
	while (argv[n]) {
	    init_env(&larg[n],argv[n]);
	    n++;
	}
	larg[n]=NULL;
    }
    else {
	larg = malloc(sizeof(char *)*4); 
	init_env(&larg[0],argv[0]);
	init_env(&larg[1],"-");
	larg[2]=NULL;
    }
    

    //
    // Finally become the user, 
    // chdir to the user home directory,
    // clean all the structures and
    // call the shell overriding current process space
    //

    setgid(gid);
    setuid(uid);
    
    ret = chdir(ptr_home_dir);	
    if (ret) {
	perror("[chrooted]chdir() ");
	return(ret);
    }

    free(home_dir);
    free(shell_path);

    ret = execve(cpwdent->pw_shell,larg,lenv);
    if (ret) { 
	perror("execve() ");
	return(errno);
    }

    //
    // not reached and if so, we are in trouble :(
    //
    
    return(-2); 
}

    
